# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
#%config Completer.use_jedi = False

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# -*- coding: utf-8 -*-
import dataiku
import pandas as pd, numpy as np
from dataiku import pandasutils as pdu
from process_mining.filtering import *
from pandas.api.types import is_numeric_dtype

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Read recipe inputs
workflow = dataiku.Dataset("workflow_parsed")
workflow_df = workflow.get_dataframe()

project = dataiku.api_client().get_project(dataiku.default_project_key())
variables = project.get_variables()

events_filters = variables['standard']['filtered_events']
min_time_filter = variables['standard']['min_time_filter']
self_transition = variables['standard']['self_transition']

add_attributes = variables['standard']['add_attributes']
attributes = variables['standard']['attributes']

# column names

case_column = variables['standard']['case']
activity_column = variables['standard']['activity']
timestamp_column = variables['standard']['timestamp']
use_end_timestamp = variables['standard']['use_end_timestamp']
end_timestamp_column = variables['standard']['end_timestamp']

if add_attributes == "select":
    for attr in attributes:
        workflow_df[attr + '_attribute'] = workflow_df[attr]

workflow_df.rename(columns={case_column: 'case',
                            activity_column: 'activity',
                            timestamp_column: 'timestamp'},
                   inplace=True)

if use_end_timestamp:
    workflow_df.rename(columns={end_timestamp_column: 'end_timestamp'},
                   inplace=True)
else:
    workflow_df['end_timestamp'] = workflow_df['timestamp']

if add_attributes == "select":
    kept_columns = ['case', 'activity', 'timestamp', 'end_timestamp'] + [attr + '_attribute' for attr in attributes]
    workflow_df = workflow_df[kept_columns]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
attributes_numeric = []

for col in attributes:
    if is_numeric_dtype(workflow_df[col + '_attribute']):
        attributes_numeric.append(1)
    else:
        attributes_numeric.append(0)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
variables['standard']['attributes_numeric'] = attributes_numeric
project.set_variables(variables)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
activity_count = workflow_df.groupby(['activity'])['case'].count().reset_index()

workflow_by_activity = dataiku.Dataset("workflow_by_activity")
workflow_by_activity.write_with_schema(activity_count)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
workflow_df = workflow_df.drop_duplicates()

workflow_df['timestamp'] = pd.to_datetime(workflow_df['timestamp'], infer_datetime_format=True, utc=True)
workflow_df['epoch_timestamp'] = pd.to_datetime(workflow_df['timestamp']).astype(int) / 10**9
workflow_df = workflow_df.sort_values(['case', 'epoch_timestamp'])
workflow_df['diff_time'] = workflow_df.groupby(['case'])['epoch_timestamp'].diff().fillna(10000)
workflow_df['previous_activity'] = workflow_df.groupby(['case'])['activity'].shift(1)

filtered_workflow = filter_activities(workflow_df, events_filters)
filtered_workflow = filter_time_before(filtered_workflow, min_time_filter)
if self_transition:
    len_workflow = len(filtered_workflow)
    last_len_workflow = 0
    while len_workflow != last_len_workflow:
        filtered_workflow['previous_activity'] = filtered_workflow.groupby(['case'])['activity'].shift(1)
        filtered_workflow = filter_self_transition(filtered_workflow)
        last_len_workflow = len_workflow
        len_workflow = len(filtered_workflow)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
case_activity_count = filtered_workflow.groupby('case')['activity'].count().reset_index()
unique_activity_case = case_activity_count[case_activity_count['activity']==1]['case']

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
filtered_workflow = filtered_workflow[~filtered_workflow['case'].isin(unique_activity_case)]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
filtered_workflow['end_timestamp'] = pd.to_datetime(filtered_workflow['end_timestamp'], infer_datetime_format=True, utc=True)
filtered_workflow['epoch_end_timestamp'] = pd.to_datetime(filtered_workflow['end_timestamp']).astype(int) / 10**9

aggs = { attribute + '_attribute': 'first' for attribute in attributes}
aggs['timestamp'] = 'min'
aggs['epoch_timestamp'] = 'min'
aggs['end_timestamp'] = 'max'
aggs['epoch_end_timestamp'] = 'max'

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
start_ends = filtered_workflow.groupby('case').agg(aggs).reset_index()

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
start_ends.columns = ['case'] + [attribute + '_attribute' for attribute in attributes] + ['min', 'epoch_min', 'max', 'epoch_max']

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
starts = start_ends[['case'] + [attribute + '_attribute' for attribute in attributes] + ['min', 'epoch_min']]
starts['activity'] = 'START'
starts.columns = ['case'] + [attribute + '_attribute' for attribute in attributes] + ['timestamp', 'epoch_timestamp', 'activity']
starts['timestamp'] = starts['timestamp'] - pd.Timedelta(seconds=1)
starts['epoch_timestamp'] = starts['epoch_timestamp'] - 1
starts['end_timestamp'] = starts['timestamp']
starts['epoch_end_timestamp'] = starts['epoch_timestamp']

ends = start_ends[['case'] + [attribute + '_attribute' for attribute in attributes] + ['max', 'epoch_max']]
ends['activity'] = 'END'
ends.columns = ['case'] + [attribute + '_attribute' for attribute in attributes] + ['timestamp', 'epoch_timestamp', 'activity']
ends['timestamp'] = ends['timestamp'] + pd.Timedelta(seconds=1)
ends['epoch_timestamp'] = ends['epoch_timestamp'] + 1
ends['end_timestamp'] = ends['timestamp']
ends['epoch_end_timestamp'] = ends['epoch_timestamp']


filtered_workflow = pd.concat([starts, filtered_workflow, ends])

filtered_workflow['process_time'] = (filtered_workflow['end_timestamp'] - filtered_workflow['timestamp']).astype('timedelta64[s]')

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
attributes_columns = [col for col in filtered_workflow.columns if col not in ['case', 'activity', 'timestamp', 'end_timestamp']]
categorical_attributes_columns =  [col for col in attributes_columns if not is_numeric_dtype(filtered_workflow[col])]

unique_attributes = pd.DataFrame()
for col in categorical_attributes_columns:
    unique_values = filtered_workflow[col].unique()
    unique_attributes = unique_attributes.append(pd.DataFrame({'attribute': [col for _ in unique_values], 'values': unique_values}))

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
unique_attributes_dataset = dataiku.Dataset("unique_attributes_values")
unique_attributes_dataset.write_with_schema(unique_attributes)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
numeric_attributes_columns = [col for col in attributes_columns if is_numeric_dtype(filtered_workflow[col])]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
numeric_attributes_scales = pd.DataFrame()
for col in numeric_attributes_columns:
    scale = np.nanpercentile(filtered_workflow[col], range(101))
    numeric_attributes_scales = numeric_attributes_scales.append(pd.DataFrame({'attribute': [col for _ in scale], 'scale': scale}))

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
numeric_attributes_scales_dataset = dataiku.Dataset("numeric_attributes_scales")
numeric_attributes_scales_dataset.write_with_schema(numeric_attributes_scales)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Write recipe outputs
workflow_clean = dataiku.Dataset("workflow_clean")
workflow_clean.write_with_schema(filtered_workflow)