import dataiku
from datetime import datetime
import dash
from dash import html, dcc
import dash_bootstrap_components as dbc
from dash import dash_table
import dash_daq as daq
from dash.dependencies import Input, Output, State
import dash_interactive_graphviz
from process_mining.graph_creation import create_graph
from process_mining.process_mining import mine_process
from process_mining.filtering import *
from webapp.selectors import contained_intersecting, filter_title, start_step, end_step, case_performance_slider, filter_element, single_datetime_filter_picker, filter_element_toggle, filter_element_range
from webapp.attribute_selectors import attribute_filters
from webapp.process_selector import process_input_name, saved_processes_dropdown, process_dropdown, process_header_menu
from webapp.filters import time_before_filter_object, dropdown_excluded_activities, remove_self_transition
from webapp.variants import variants_container, variants_filter_object
from webapp.selected_elements import selected_node_window, selected_edge_window, selected_node_card, empty_card, selected_edge_card, empty_card_individual, selected_node_individual_card, selected_edge_individual_card
from webapp.styles import *
from webapp.legend import legend_object, legend_object_individual
from webapp.filter_processing import apply_filters, base_filters, additional_filters
from webapp.explanation import process_mining_explanation, conformance_explanation
from webapp.header import filter_header, filter_subtitle, header_conformance_children, header_children, exclude_subtitle
from webapp.save_modal import save_modal
from webapp.process_management import save_process_in_folder, delete_process_in_folder
from webapp.layout import layout_discovery, layout_conformance, layout_individual
from data_management import get_managed_folder_id_with_folder_name, write_pickle_in_dss_folder, read_pickle_from_dss_folder
from webapp.formatting import human_format, formatted_human_readable, start_week
import pm4py
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
import pickle
import ast
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import numpy as np
from random import sample
from dash.exceptions import PreventUpdate
import logging
import threading
import time
from cairosvg import svg2png
import pydot
from dataiku import insights
from flask import send_from_directory

def get_assets_folder():
    result = None
    for path_ in os.environ.get("PYTHONPATH", "").split(os.pathsep):
        if os.path.exists(os.path.join(path_,"assets")):
            result = os.path.join(path_,"assets")
    return result

@app.server.route("/api/<filename>")
def get_file(filename):
    return send_from_directory(get_assets_folder(), filename)


timeframe_df = None
workflow_df = None
total_cases = None
start_end_df = None
case_performance_df = None
activities_count_df = None
activity_count_df = None
transitions_df = None
case_performance_scale = None
attributes_columns = None
attribute_scales = None
attributes_values_df = None
graph = None
individual_graph = None
variants_count = None

def load_small_data():
    global timeframe_df
    global start_end_df
    global attributes_values_df
    global attribute_scales
    global activity_count_df
    global case_performance_df
    global case_performance_scale
    
    logging.info("Current step: timeframe")
    timeframe = dataiku.Dataset('timeframe')
    timeframe_df = timeframe.get_dataframe()
    
    logging.info("Current step: start_end")
    start_end = dataiku.Dataset("start_end")
    start_end_df = start_end.get_dataframe()
    
    try:
        logging.info("Current step: attributes_values")
        attributes_values = dataiku.Dataset("unique_attributes_values")
        attributes_values_df = attributes_values.get_dataframe()
    except:
        attributes_values_df = pd.DataFrame()
    
    try:
        logging.info("Current step: attribute_scales")
        attributes_scales = dataiku.Dataset("numeric_attributes_scales")
        attributes_scales_df = attributes_scales.get_dataframe()
        numeric_attributes = attributes_scales_df['attribute'].unique()
        attribute_scales = {col.replace('_attribute', ''): sorted(attributes_scales_df[attributes_scales_df['attribute']==col]['scale'].to_list()) for col in numeric_attributes}
    except:
        attribute_scales = dict()
    
    logging.info("Current step: workflow_by_activity")
    activity_count = dataiku.Dataset("workflow_by_activity")
    activity_count_df = activity_count.get_dataframe()
    
    logging.info("Current step: workflow_case_performance")
    case_performance = dataiku.Dataset("workflow_case_performance")
    case_performance_df = case_performance.get_dataframe()
    
    logging.info("Current step: case_performance_scale")
    case_performance_scale = np.percentile(case_performance_df['case_performance'], range(101))

    
def load_data():
    global workflow_df
    global total_cases
    global activities_count_df
    global transitions_df
    
    logging.info("Current step: workflow_clean")
    workflow = dataiku.Dataset("workflow_clean")
    logging.info("Current step: dataframe")
    workflow_df = workflow.get_dataframe()
    workflow_df = workflow_df.sort_values(by=['timestamp'])
    
    logging.info("Current step: case")
    total_cases = len(workflow_df['case'].unique())
    
    logging.info("Current step: activity_count")
    activities_count = dataiku.Dataset("activity_count")
    activities_count_df = activities_count.get_dataframe()
    
    logging.info("Current step: transitions_frequency_share_prepared")
    transitions_frequency_share = dataiku.Dataset("transitions_frequency_share_prepared")
    transitions_df = transitions_frequency_share.get_dataframe()
    logging.info("Loading: finished")


logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.DEBUG, datefmt="%H:%M:%S")
logging.info("Main    : before creating thread")

loading_thread = threading.Thread(target=load_data)
logging.info("Main    : before running loading data thread")
loading_thread.start()

small_loading_thread = threading.Thread(target=load_small_data)
logging.info("Main    : before running loading data thread")
small_loading_thread.start()

palette = '#D5D9D9', '#3075AE', '#ff7e0b'

old_current_process = None
current_process = None
final_process = None
individual_data = None

app.config.external_stylesheets = [dbc.themes.BOOTSTRAP, "https://use.fontawesome.com/releases/v5.13.0/css/all.css", "api/app.css"]

client = dataiku.api_client()
project_key = dataiku.default_project_key()
project = client.get_project(project_key)
variables = project.get_variables()

events_filters = variables['standard']['filtered_events']
min_time_filter = variables['standard']['min_time_filter']
self_transition = variables['standard']['self_transition']
attributes_columns = variables['standard']['attributes']
attributes_numeric = variables['standard']['attributes_numeric']
use_end_timestamp = variables['standard']['use_end_timestamp']
end_time_column = variables['standard']['end_timestamp']

palette = ['#BDD8ED', '#3075AE', '#4F934F']

@app.callback(Output('nb-cases-info', 'children'),
              Output('graph-legend', 'children'),
              Output('graph-workflow', 'children'),
              Output('unique-nodes', 'data'),
              Output('old-variants-filter', 'data'),
              Output('gradient-legend', 'style'),
              State('graph-workflow', 'children'),
              State('unique-nodes', 'data'),
              State('old-variants-filter', 'data'),
              Input('variants-filter', 'value'),
              Input('time-freq-state', 'value'),
              Input('tab-state', 'value'))
def render_content(old_graph, unique_nodes, old_nb_variants, 
                   nb_variants, frequency_time, tab):
    global old_current_process
    global final_process
    global graph
    
    ctx = dash.callback_context
    if old_current_process is not None:
        if old_nb_variants==nb_variants and old_current_process.equals(current_process) and ctx.triggered[0]['prop_id'] == 'variants-filter.value' and old_graph is not None:
            raise PreventUpdate
    if current_process is not None:
        filtered_workflow = filter_workflow(current_process, nb_variants, variants_count)
        filtered_cases = len(current_process['case'].unique())
        displayed_cases = len(filtered_workflow['case'].unique())
        
        if filtered_cases == 0:
            return [html.P('No cases filtered, try releasing the constraints', style={'font-size': '12px', 'margin': '1px'})], legend_object(1, 1, frequency_time == 'time'), dbc.Col([None], style={'width': '600px'}), None, 0, {'height': '50%',
               'width': '100%',
               'margin': 'auto',
               'background': 'linear-gradient(to right, #F5F0F9, #A47BC7)',
               'padding': 0}
        
        if frequency_time == 'frequency':
            workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', transition='frequency')
        else:
            if use_end_timestamp:
                workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', 'end_timestamp', transition='performance')
            else:
                workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', transition='performance')
        graph, unique_nodes, max_count = create_graph(workflow_mined, start_end, activities, transition=frequency_time)
        graph_object = dash_interactive_graphviz.DashInteractiveGraphviz(
            id="filtered-workflow",
            dot_source=graph,
            style={'width': '75%', 'height': '85%',
                  'cursor': 'pointer'}
        )
        final_process = filtered_workflow
        old_current_process = current_process
        if frequency_time=='time':
            legend_style = {'height': '50%',
               'width': '100%',
               'margin': 'auto',
               'background': 'linear-gradient(to right, #F5F0F9, #A47BC7)',
               'padding': 0}
        else:
            legend_style = {'height': '50%',
               'width': '100%',
               'margin': 'auto',
               'background': 'linear-gradient(to right, #F1F8FD, #7DBEEA)',
               'padding': 0}
        return [html.P(human_format(displayed_cases) + ' cases displayed on ' + human_format(total_cases) + ' total', style={'font-size': '12px', 'margin': '1px'})], legend_object(1 if frequency_time=='frequency' else 0, max_count, frequency_time), dbc.Col([graph_object], style={'width': '600px'}), unique_nodes, nb_variants, legend_style
    else:
        raise PreventUpdate

@app.callback(Output('filters', 'data'),
              Output('variants-container', 'children'),
              State('old-variants-filter', 'data'),
              Input('tab-state', 'value'),
              Input('start-date-filter', 'value'),
              Input('end-date-filter', 'value'),
              Input('contained-intersecting', 'on'),
              Input('start-step', 'value'),
              Input('end-step', 'value'),
              Input('case-performance-filter', 'value'),
              Input('excluded-activities-dropdown', 'value'),
              Input('time-before-activity-filter', 'value'),
              Input('self-transition', 'on'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns])
def render_content(variants, tab, start_date, end_date, contained, 
                   start_steps, end_steps, case_performance, 
                   excluded_activities, time_before_activity, self_transition,
                   *attributes):
    logging.info('Current step: Joining the thread')
    loading_thread.join()
    logging.info('Current step: Thread joined')
    global current_process
    global variants_count
    ctx = dash.callback_context
    if ctx.triggered[0]["prop_id"] == 'tab-state.value':
        if tab != 'discovery':
            return None, variants_filter_object(tab=tab)
    
    filtered_workflow = apply_filters(workflow_df, start_date, end_date, contained, 
                                       start_steps, end_steps, case_performance, case_performance_scale, 
                                       attributes_columns, attribute_scales,
                                       excluded_activities, time_before_activity, self_transition,
                                       *attributes)
    
    filters = {'start_date': start_date, 'end_date': end_date,
              'contained': contained, 
               'start_steps': start_steps, 'end_steps': end_steps, 
               'case_performance': case_performance, 
               'excluded_activities': excluded_activities, 
               'time_before_activity': time_before_activity, 
               'self_transition': self_transition}
    
    for i, attribute in enumerate(attributes):
        filters[attributes_columns[i]] = attribute
        
    variants_count = case_statistics.get_variant_statistics(filtered_workflow,
                                          parameters={case_statistics.Parameters.CASE_ID_KEY: "case",
                                                      case_statistics.Parameters.ACTIVITY_KEY: "activity",
                                                      case_statistics.Parameters.TIMESTAMP_KEY: "timestamp"})
    variants_count = sorted(variants_count, key=lambda x: x['case'], reverse=True)
    
    if variants is None:
        variants = 10
    
    current_process = filtered_workflow
    return filters, variants_filter_object(variants, len(variants_count))

@app.callback(Output('selected', 'children'),
              Output('selected', 'style'),
              Input('filtered-workflow', 'selected_node'),
              Input('filtered-workflow', 'selected_edge'),
              Input('card-close-button', 'n_clicks'),
              State('unique-nodes', 'data'))
def output_selected(selected_node, selected_edge, n_close, data):
    ctx = dash.callback_context

    if not ctx.triggered:
        raise PreventUpdate
    
    button = ctx.triggered[0]["prop_id"]
    if button == 'card-close-button.n_clicks':
        return empty_card(), {'position': 'absolute',
                              'z-index': '999999',
                              'display': 'None'}
    if selected_node is not None:
        node = data[int(selected_node)]
        return selected_node_card(node, selected_node, data, 
                                  activities_count_df, transitions_df, 
                                  total_cases), {
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-right': '10px',
                      'border-radius': 0,
                      'border-left': '5px solid #69B2E6',
                      'float': 'right'}
    elif selected_edge is not None:
        nodes = selected_edge.split('->')
        source = data[int(nodes[0])]
        target = data[int(nodes[1])]
        return selected_edge_card(source, target, nodes, 
                                  data, transitions_df, 
                                  total_cases), {
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-right': '10px',
                      'border-radius': 0,
                      'border-left': '5px solid #A07DC2',
                      'float': 'right'}
    else:
        return empty_card(), {'position': 'absolute',
                              'z-index': '999999',
                              'display': 'None'}

@app.callback(Output('selected-individual', 'children'),
              Output('selected-individual', 'style'),
              Input('individual-workflow', 'selected_node'),
              Input('individual-workflow', 'selected_edge'),
              Input('card-close-individual-button', 'n_clicks'),
              State('unique-nodes-individual', 'data'))
def output_selected(selected_node, selected_edge, n_close, data):
    ctx = dash.callback_context

    if not ctx.triggered:
        raise PreventUpdate
    
    button = ctx.triggered[0]["prop_id"]
    if button == 'card-close-individual-button.n_clicks':
        return empty_card_individual(), {'position': 'absolute',
      'z-index': '999999',
      'display': 'None'}
    if selected_node is not None:
        node = data[int(selected_node)]
        return selected_node_individual_card(node, selected_node, data, 
                                  activities_count_df, transitions_df, 
                                  total_cases), {
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-right': '10px',
                      'border-radius': 0,
                      'border-left': '5px solid #69B2E6',
                      'float': 'right'}
    elif selected_edge is not None:
        nodes = selected_edge.split('->')
        source = data[int(nodes[0])]
        target = data[int(nodes[1])]
        return selected_edge_individual_card(source, target, nodes, 
                                  data, transitions_df, 
                                  total_cases), {
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-right': '10px',
                      'border-radius': 0,
                      'border-left': '5px solid #A07DC2',
                      'float': 'right'}
    else:
        return empty_card_individual(), {'position': 'absolute',
      'z-index': '999999',
      'display': 'None'}

@app.callback(Output('loading', 'children'),
             Input('push-to-flow', 'n_clicks'),
             State('excluded-activities-dropdown', 'value'),
             State('time-before-activity-filter', 'value'),
             State('self-transition', 'on'))
def push_to_flow(n_clicks, excluded_activities, time_before, self_transition):
    if n_clicks==0:
        raise PreventUpdate
    else:
        variables['standard']['filtered_events'] = excluded_activities
        variables['standard']['min_time_filter'] = time_before
        variables['standard']['self_transition'] = self_transition
        project.set_variables(variables)
        scenario = project.get_scenario("Rebuild_Flow")
        scenario.run_and_wait()

@app.callback(
    Output("save-process-modal", "is_open"),
    Output("modal-save-process", "disabled"),
    Output('saved-processes-dropdown', 'options'),
    Input("cancel-save-process", "n_clicks"),
    Input("modal-save-process", "n_clicks"),
    Input("save-process-button", "n_clicks"),
    Input("save-as-button", "n_clicks"),
    Input('process-name', 'value'),
    Input('delete-process', 'n_clicks'),
    State('saved-processes-dropdown', 'value'),
    State("save-process-modal", "is_open"),
    State('saved-processes-dropdown', 'options'),
    State('filters', 'data'))
def toggle_open_modal(cancel_button, modal_save_button, save_button, save_as_button, name, delete_process, selected_process, is_open, options, filters):
    ctx = dash.callback_context
    if len(ctx.triggered)>1:
        raise PreventUpdate
    if selected_process is not None and ctx.triggered[0]['prop_id'] == 'save-process-button.n_clicks':
        save_process_in_folder(project, selected_process, final_process, filters)
        return False, True, options
    if selected_process is not None and ctx.triggered[0]['prop_id'] == 'delete-process.n_clicks' and delete_process is not None:
        if delete_process > 0:
            #input_folder = dataiku.Folder("1rhVX912")
            #input_folder.delete_path(selected_process + '_dfg.p')
            #input_folder.delete_path(selected_process + '_filters.p')
            #input_folder.delete_path(selected_process + '_petri.p')
            delete_process_in_folder(selected_process)
            options = [opt for opt in options if opt['label'] != selected_process]
            return False, True, options
    if name is None:
        disabled = True
    else:
        if len(name) > 0:
            disabled = False
        else:
            disabled = True
    if ctx.triggered[0]['prop_id'] == 'modal-save-process.n_clicks':
        if len(name) > 0 and modal_save_button > 0:
            save_process_in_folder(project, name, final_process, filters)
            options.append({'label': name, 'value': name})
    if ctx.triggered[0]['prop_id'] == 'process-name.value':
        return is_open, disabled, options
    if cancel_button or modal_save_button or save_button or save_as_button:
        return not is_open, disabled, options
    return is_open, disabled, options

@app.callback(Output('process-header', 'children'),
              Input('tab-state', 'value'),
              State('saved-processes-value', 'data'))
def apply_or_update_process(tab, process):
    return process_dropdown(process, tab)
    
@app.callback(Output('base-filters', 'children'),
              State('tab-state', 'value'),
              Input('saved-processes-dropdown', 'value'),
              State('base-filters', 'children'))
def apply_filters_button(tab, process, old_base_filters):
    if process is None:
        raise PreventUpdate
    filters_data = read_pickle_from_dss_folder(process + '_filters.p', '1rhVX912')
    base_filters = [filter_element('Traces starting at', single_datetime_filter_picker(filters_data['start_date'], True)),
                     filter_element('Traces ending at', single_datetime_filter_picker(filters_data['end_date'], False)),
                     filter_element_toggle('Traces must be strictly within range', contained_intersecting(filters_data['contained'])),
                     filter_element('Starting activities', start_step(start_end_df, filters_data['start_steps'])),
                     filter_element('Ending activities', end_step(start_end_df, filters_data['end_steps'])),
                     filter_element_range('Traces duration', case_performance_slider(filters_data['case_performance'])),
                     html.Div(attribute_filters(attributes_values_df, attributes_columns, attributes_numeric, attribute_scales, {attribute: filters_data[attribute] for attribute in attributes_columns}), id='attribute-filters')]
    return base_filters

@app.callback(Output('min-traces-duration', 'children'),
              Output('max-traces-duration', 'children'),
              Input('case-performance-filter', 'value'))
def output_duration_bounds(case_performance):
    return html.P(formatted_human_readable(case_performance_scale[case_performance[0]])), html.P(formatted_human_readable(case_performance_scale[case_performance[1]]))

@app.callback([Output('min-attribute-filter-{}'.format(attributes_column), 'children') for i, attributes_column in enumerate(attributes_columns) if attributes_numeric[i]==1],
              [Output('max-attribute-filter-{}'.format(attributes_column), 'children') for i, attributes_column in enumerate(attributes_columns) if attributes_numeric[i]==1],
              Output('dummy-output', 'data'),
              Input('dummy-input', 'data'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for i, attributes_column in enumerate(attributes_columns) if attributes_numeric[i]==1])
def output_duration_bounds(dummy_input, values):
    if len(attribute_scales) == 0:
        return [dummy_input]
    columns = [attributes_column for i, attributes_column in enumerate(attributes_columns) if attributes_numeric[i]==1]
    if len(columns) == 1:
        values = [values]
    return [html.P(human_format(attribute_scales[columns[i]][value[0]])) for i, value in enumerate(values)], [html.P(human_format(attribute_scales[columns[i]][value[1]])) for i, value in enumerate(values)], dummy_input

@app.callback(Output('individual-processes', 'children'),
              Input('saved-processes-dropdown', 'value'),
              Input('conformance-button', 'n_clicks'),
              Input('start-date-filter', 'value'),
              Input('end-date-filter', 'value'),
              Input('contained-intersecting', 'on'),
              Input('start-step', 'value'),
              Input('end-step', 'value'),
              Input('case-performance-filter', 'value'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns])
def individual_processes_table(process, conformance_button, 
                               start_date, end_date, contained, 
                               start_steps, end_steps, case_performance, 
                               *attributes):
    ctx = dash.callback_context
    if ctx.triggered[0]['prop_id'] == 'conformance-button.n_clicks' and conformance_button is not None:
        if conformance_button > 0:
            scenario = project.get_scenario("RUNCONFORMANCECHECKS")
            scenario.run_and_wait()
    
    global individual_data
    filtered_workflow = apply_filters(workflow_df, start_date, end_date, contained, 
                                               start_steps, end_steps, case_performance, case_performance_scale,
                                               attributes_columns, attribute_scales,
                                               events_filters, min_time_filter, self_transition,
                                               *attributes)
    case_ids = [str(x) for x in filtered_workflow['case'].unique()]
    conformance_checks = dataiku.Dataset('process_conformance_checks')
    try:
        conformance_checks_df = conformance_checks.get_dataframe()
    except:
        return dbc.Container([html.P('Conformance checks compare traces against a reference process.'),
                                 html.P('You can compare traces against processes, checking the correspondance between the two process.')],
                                style={'font-size': '18px',
                                       'text-align': 'center',
                                       'color': '#666666',
                                       'margin-top': '150px',
                                       'font-weight': '400'})
    if process not in conformance_checks_df['reference_process'].unique() or process is None:
        return dbc.Container([html.P('Conformance checks compare traces against a reference process.'),
                                 html.P('You can compare traces against processes, checking the correspondance between the two process.')],
                                style={'font-size': '18px',
                                       'text-align': 'center',
                                       'color': '#666666',
                                       'margin-top': '150px',
                                       'font-weight': '400'})
    checks = conformance_checks_df[conformance_checks_df['case'].isin(case_ids)]
    checks = checks[checks['reference_process']==process]
    table_columns = ['case', 'trace_fitness', 'timestamp'] + [col + '_attribute' for col in attributes_columns]
    table_columns_names = ['Case', 'Fitness', 'Start Date'] + attributes_columns
    individual_data = [{col: checks[col].iloc[i] for col in table_columns}
                  for i in range(len(checks))]
    individual_processes_table = dash_table.DataTable(
            id='individual-processes-table',
            columns=(
                    [{'id': table_columns[i], 'name': table_columns_names[i]}
                     for i in range(len(table_columns))]
                 ),
            data=individual_data,
        style_cell={'textAlign': 'left',
                    'font-family': 'Helvetica Neue'},
        active_cell={'column': 0, 'row': 0},
        sort_action='native',
        style_table={'height': '20em', 'overflowY': 'auto', 'cursor': 'pointer'},
        style_data_conditional=[
            {
                'if': {
                    'state': 'active'
                },
                'backgroundColor': '#eef5ff',
                "border": "1px solid #288cfe"
            },
            {
                'if': {
                    'state': 'selected'
                },
                'backgroundColor': '#eef5ff',
                "border": "1px solid #288cfe"
            },
        ]
    )
    return individual_processes_table

@app.callback(Output('graphs', 'children'),
              Input('tab-state', 'value'),
              Input('conformance-button', 'n_clicks'),
              State('saved-processes-dropdown', 'value'),
              Input('conformance-granularity', 'value'),
              Input('start-date-filter', 'value'),
              Input('end-date-filter', 'value'),
              Input('contained-intersecting', 'on'),
              Input('start-step', 'value'),
              Input('end-step', 'value'),
              Input('case-performance-filter', 'value'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns])
def make_figures(tab, n_clicks, process, conformance_granularity,
                 start_date, end_date, contained, 
                 start_steps, end_steps, case_performance, 
                 *attributes):
    ctx = dash.callback_context
    if ctx.triggered[0]['prop_id'] == 'conformance-button.n_clicks' and n_clicks is not None:
        if n_clicks > 0:
            scenario = project.get_scenario("RUNCONFORMANCECHECKS")
            scenario.run_and_wait()
    conformance_checks = dataiku.Dataset("process_conformance_checks")
    try:
        conformance_checks_df = conformance_checks.get_dataframe()
    except:
        return dbc.Container([html.P('Conformance checks compare traces against a reference process.'),
                                 html.P('You can compare traces against processes, checking the correspondance between the two process.')],
                                style={'font-size': '18px',
                                       'text-align': 'center',
                                       'color': '#666666',
                                       'margin-top': '150px',
                                       'font-weight': '400'})
    if process is None:
        return dbc.Container([html.P('Conformance checks compare traces against a reference process.'),
                                 html.P('You can compare traces against processes, checking the correspondance between the two process.')],
                                style={'font-size': '18px',
                                       'text-align': 'center',
                                       'color': '#666666',
                                       'margin-top': '150px',
                                       'font-weight': '400'})
    if process not in conformance_checks_df['reference_process'].unique():
        return dbc.Container([html.P('Conformance checks compare traces against a reference process.'),
                                 html.P('You can compare traces against processes, checking the correspondance between the two process.')],
                                style={'font-size': '18px',
                                       'text-align': 'center',
                                       'color': '#666666',
                                       'margin-top': '150px',
                                       'font-weight': '400'})
    filtered_workflow = apply_filters(workflow_df, start_date, end_date, contained, 
                                           start_steps, end_steps, case_performance, case_performance_scale,
                                           attributes_columns, attribute_scales,
                                           events_filters, min_time_filter, self_transition,
                                           *attributes)
    case_ids = filtered_workflow['case'].unique()
    checks = conformance_checks_df[conformance_checks_df['reference_process']==process]
    checks = checks[checks['case'].isin(case_ids)]
    if conformance_granularity == 'monthly':
        checks['timestamp'] = [datetime(d.year, d.month, 1) for d in checks['timestamp']]
        fitness_date = checks.groupby(['timestamp'])[['trace_fitness', 'trace_is_fit']].mean().reset_index()
    elif conformance_granularity == 'weekly':
        checks['timestamp'] = [start_week(d) for d in checks['timestamp']]
        fitness_date = checks.groupby(['timestamp'])[['trace_fitness', 'trace_is_fit']].mean().reset_index()
    else:
        checks['timestamp'] = [datetime(d.year, d.month, d.day) for d in checks['timestamp']]
        fitness_date = checks.groupby(['timestamp'])[['trace_fitness', 'trace_is_fit']].mean().reset_index()
    global_fitness = checks['trace_fitness'].mean()
    global_fit_traces = checks['trace_is_fit'].mean()
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=fitness_date["timestamp"], y=fitness_date["trace_fitness"], 
                  name='Fitness Evolution (Average Fitness ' + str(round(global_fitness, 3)) + ')',
                  line=dict(color=palette[0])))
    fig.add_trace(go.Scatter(x=fitness_date["timestamp"], y=fitness_date["trace_is_fit"], 
                  name='Share of Fit Traces (Average Share ' + str(round(global_fit_traces, 3)) + ')',
                  line=dict(color=palette[1])))
    fig.update_layout(plot_bgcolor='rgba(0,0,0,0)')
    fig.layout.yaxis.gridcolor = '#D7DBDE'
    fig.layout.margin.t = 10
    fig.update_layout(legend=dict(
    orientation="h",
    yanchor="top",
    y=-0.05,
    xanchor="right",
    x=1))
    return dbc.Col(dcc.Graph(figure=fig, 
                             id='conformance-checks-evolution',
                            style={'margin': 0}))

@app.callback(Output('filter-subtitle', 'children'),
              Input('tab-state', 'value'))
def update_title(tab):
    if tab == 'discovery':
        title = 'Selectors'
    elif tab == 'conformance':
        title = 'to traces'
    else:
        title = 'to individual traces'
    return html.P(title, style={'font-weight': 'bold', 
                                'text-align': 'left', 
                                'margin-bottom': 0})

@app.callback(Output('filter-header', 'children'),
              Input('tab-state', 'value'))
def update_title(tab):
    if tab == 'discovery':
        return header_children
    else:
        return header_conformance_children

@app.callback(
    Output("save-modal", "is_open"),
    Output("modal-save-viz", "disabled"),
    Input("cancel-save", "n_clicks"),
    Input("modal-save-viz", "n_clicks"),
    Input("save-viz", "n_clicks"),
    Input('viz-name', 'value'),
    State("save-modal", "is_open"),
    State('tab-state', 'value'))
def toggle_open_modal(cancel_button, modal_save_button, save_button, name, is_open, tab):
    ctx = dash.callback_context
    if name is None:
        disabled = True
    else:
        if len(name) > 0:
            disabled = False
        else:
            disabled = True
    if ctx.triggered[0]['prop_id'] == 'modal-save-viz.n_clicks':
        if len(name) > 0 and modal_save_button > 0:
            if tab == 'discovery':
                dot_graph = pydot.graph_from_dot_data(graph)
            else:
                dot_graph = pydot.graph_from_dot_data(individual_graph)
            folder_id = 'LsTJHTYd'
            svg = dot_graph[0].create_svg()
            png = svg2png(svg)
            dss_folder = dataiku.Folder(folder_id)
            with dss_folder.get_writer(name + '.png') as w:
                w.write(png)
            insights.save_data(name, png, 'image/png')
    if ctx.triggered[0]['prop_id'] == 'viz-name.value':
        return is_open, disabled
    if cancel_button or modal_save_button or save_button:
        return not is_open, disabled
    return is_open, disabled

@app.callback(Output('individual-graph', 'children'),
              Output('graph-legend-individual', 'children'),
              Output('unique-nodes-individual', 'data'),
              Output('gradient-legend-individual', 'style'),
              Output('individual-graph-top-bar', 'style'),
              State('individual-processes-table', 'derived_virtual_indices'),
              Input('individual-processes-table', 'active_cell'),
              Input('time-freq-individual-state', 'value'))
def render_graph(indices, cell, frequency_time):
    global individual_graph
    if indices is not None and cell is not None:
        case = individual_data[indices[cell['row']]]['case']
        filtered_workflow = workflow_df[workflow_df['case']==case]
        workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', end_time=('end_timestamp' if use_end_timestamp else None), transition=('performance' if frequency_time=='time' else 'frequency'))
        individual_graph, unique_nodes, max_count = create_graph(workflow_mined, start_end, activities, transition=frequency_time)
        graph_object = dash_interactive_graphviz.DashInteractiveGraphviz(
            id="individual-workflow",
            dot_source=individual_graph,
            style={'width': '75%', 'height': '67%',
                  'cursor': 'pointer'}
        )
        if frequency_time=='time':
            legend_style = {'height': '50%',
               'width': '100%',
               'margin': 'auto',
               'background': 'linear-gradient(to right, #F5F0F9, #A47BC7)',
               'padding': 0}
        else:
            legend_style = {'height': '50%',
               'width': '100%',
               'margin': 'auto',
               'background': 'linear-gradient(to right, #F1F8FD, #7DBEEA)',
               'padding': 0}
        return graph_object, legend_object_individual(1 if frequency_time=='frequency' else 0, max_count, frequency_time), unique_nodes, legend_style, {'display': 'flex', 'margin-left': '0px'}
    else:
        raise PreventUpdate


@app.callback(Output('tab-content', 'children'),
              Input('tab-state', 'value'),
              Input('list-chart-state', 'value'),
              Input('conformance-button', 'n_clicks'),
              State('saved-processes-dropdown', 'value'))
def render_content(tab, line_chart_tab, check_conformance, process):
    ctx = dash.callback_context
    if ctx.triggered[0]['prop_id'] == 'conformance-button.n_clicks' and check_conformance is not None:
        if check_conformance > 0:
            scenario = project.get_scenario("RUNCONFORMANCECHECKS")
            scenario.run_and_wait()
    if tab == 'discovery':
        return layout_discovery(project_key)
    else:
        if process is None:
            return dbc.Container([html.P('Conformance checks compare traces against a reference process.'),
                                 html.P('You can compare traces against processes, checking the correspondance between the two process.')],
                                style={'font-size': '18px',
                                       'text-align': 'center',
                                       'color': '#666666',
                                       'margin-top': '150px',
                                       'font-weight': '400'})
        else:
            if line_chart_tab == 'chart':
                return layout_conformance()
            else:
                return layout_individual(project_key)

@app.callback(Output('additional-filters-zone', 'children'),
              Input('tab-state', 'value'))
def render_content(tab):
    if tab == 'discovery':
        return [exclude_subtitle(),
                dbc.Collapse(
                            additional_filters(activity_count_df, events_filters, min_time_filter, self_transition),
                            id="filters-collapse",
                            is_open=True,
                     )]
    else:
        return None

@app.callback(Output('radio-buttons', 'children'),
              Output('tab-state', 'value'),
             Input('discovery-tab', 'n_clicks'),
             Input('conformance-tab', 'n_clicks'),
             State('tab-state', 'value'))
def update_buttons(discovery_button, conformance_button, tab_state):
    ctx = dash.callback_context
    if discovery_button is None and conformance_button is None:
        buttons = [html.Button('Discovery', id='discovery-tab', style=tab_button_style_active),
             html.Button('Conformance', id='conformance-tab', style=tab_button_style)]
        return buttons, 'discovery'
    if ctx.triggered[0]['prop_id']=='discovery-tab.n_clicks':
        if tab_state=='discovery':
            raise PreventUpdate
        else:
            buttons = [html.Button('Discovery', id='discovery-tab', style=tab_button_style_active),
             html.Button('Conformance', id='conformance-tab', style=tab_button_style)]
            return buttons, 'discovery'
    else:
        if tab_state=='conformance':
            raise PreventUpdate
        else:
            buttons = [html.Button('Discovery', id='discovery-tab', style=tab_button_style),
             html.Button('Conformance', id='conformance-tab', style=tab_button_style_active)]
            return buttons, 'conformance'


@app.callback(Output('radio-buttons-time-freq', 'children'),
              Output('time-freq-state', 'value'),
             Input('freq-tab', 'n_clicks'),
             Input('time-tab', 'n_clicks'),
             State('time-freq-state', 'value'))
def update_buttons(freq_button, time_button, tab_state):
    ctx = dash.callback_context
    if freq_button is None and time_button is None:
        left_style = tab_button_style_freq_time_active.copy()
        left_style['border-right'] = '1px solid #666666'
        buttons = [html.Button('Frequency', id='freq-tab', style=left_style),
             html.Button('Time', id='time-tab', style=tab_button_style_freq_time)]
        return buttons, 'frequency'
    if ctx.triggered[0]['prop_id']=='freq-tab.n_clicks':
        if tab_state=='frequency':
            raise PreventUpdate
        else:
            left_style = tab_button_style_freq_time_active.copy()
            left_style['border-right'] = '1px solid #666666'
            buttons = [html.Button('Frequency', id='freq-tab', style=left_style),
             html.Button('Time', id='time-tab', style=tab_button_style_freq_time)]
            return buttons, 'frequency'
    else:
        if tab_state=='time':
            raise PreventUpdate
        else:
            left_style = tab_button_style_freq_time.copy()
            left_style['border-right'] = '1px solid #666666'
            buttons = [html.Button('Frequency', id='freq-tab', style=left_style),
             html.Button('Time', id='time-tab', style=tab_button_style_freq_time_active)]
            return buttons, 'time'

@app.callback(Output('radio-buttons-time-freq-individual', 'children'),
              Output('time-freq-individual-state', 'value'),
             Input('freq-tab-individual', 'n_clicks'),
             Input('time-tab-individual', 'n_clicks'),
             State('time-freq-individual-state', 'value'))
def update_buttons(freq_button, time_button, tab_state):
    ctx = dash.callback_context
    if freq_button is None and time_button is None:
        left_style = tab_button_style_freq_time_active.copy()
        left_style['border-right'] = '1px solid #666666'
        buttons = [html.Button('Frequency', id='freq-tab-individual', style=left_style),
             html.Button('Time', id='time-tab-individual', style=tab_button_style_freq_time)]
        return buttons, 'frequency'
    if ctx.triggered[0]['prop_id']=='freq-tab-individual.n_clicks':
        if tab_state=='frequency':
            raise PreventUpdate
        else:
            left_style = tab_button_style_freq_time_active.copy()
            left_style['border-right'] = '1px solid #666666'
            buttons = [html.Button('Frequency', id='freq-tab-individual', style=left_style),
             html.Button('Time', id='time-tab-individual', style=tab_button_style_freq_time)]
            return buttons, 'frequency'
    else:
        if tab_state=='time':
            raise PreventUpdate
        else:
            left_style = tab_button_style_freq_time.copy()
            left_style['border-right'] = '1px solid #666666'
            buttons = [html.Button('Frequency', id='freq-tab-individual', style=left_style),
             html.Button('Time', id='time-tab-individual', style=tab_button_style_freq_time_active)]
            return buttons, 'time'

        
@app.callback(Output('list-chart-radio-buttons', 'children'),
              Output('list-chart-state', 'value'),
             Input('conformance-graph-tab', 'n_clicks'),
             Input('conformance-table-tab', 'n_clicks'),
             State('list-chart-state', 'value'))
def update_buttons(graph_button, table_button, tab_state):
    ctx = dash.callback_context
    if graph_button is None and table_button is None:
        if tab_state is None:
            buttons = [html.Button(html.I(className="fas fa-list"), id='conformance-table-tab', style=tab_button_style_list_chart_active),
                 html.Button(html.I(className="fas fa-chart-line"), id='conformance-graph-tab', style=tab_button_style_list_chart)]
            return buttons, 'table'
        else:
            if tab_state == 'table':
                buttons = [html.Button(html.I(className="fas fa-list"), id='conformance-table-tab', style=tab_button_style_list_chart_active),
                 html.Button(html.I(className="fas fa-chart-line"), id='conformance-graph-tab', style=tab_button_style_list_chart)]
                return buttons, 'table'
            else:
                buttons = [html.Button(html.I(className="fas fa-list"), id='conformance-table-tab', style=tab_button_style_list_chart),
             html.Button(html.I(className="fas fa-chart-line"), id='conformance-graph-tab', style=tab_button_style_list_chart_active)]
                return buttons, 'chart'
    if ctx.triggered[0]['prop_id']=='conformance-table-tab.n_clicks':
        if tab_state=='table':
            raise PreventUpdate
        else:
            buttons = [html.Button(html.I(className="fas fa-list"), id='conformance-table-tab', style=tab_button_style_list_chart_active),
             html.Button(html.I(className="fas fa-chart-line"), id='conformance-graph-tab', style=tab_button_style_list_chart)]
            return buttons, 'table'
    else:
        if tab_state=='chart':
            raise PreventUpdate
        else:
            buttons = [html.Button(html.I(className="fas fa-list"), id='conformance-table-tab', style=tab_button_style_list_chart),
             html.Button(html.I(className="fas fa-chart-line"), id='conformance-graph-tab', style=tab_button_style_list_chart_active)]
            return buttons, 'chart'

@app.callback(
    Output("selectors-collapse", "is_open"),
    Output("expand-selector-button", 'children'),
    Input("expand-selector-button", "n_clicks"),
    State("selectors-collapse", "is_open"))
def toggle_collapse(n, is_open):
    if n:
        if is_open:
            return not is_open, [html.I(className="fas fa-chevron-right")]
        else:
            return not is_open, [html.I(className="fas fa-chevron-down")]
    if is_open:
        return is_open, [html.I(className="fas fa-chevron-down")]
    else:
        return is_open, [html.I(className="fas fa-chevron-right")]

@app.callback(
    Output("filters-collapse", "is_open"),
    Output("expand-filter-button", 'children'),
    Input("expand-filter-button", "n_clicks"),
    State("filters-collapse", "is_open"))
def toggle_expand_collapse(n, is_open):
    if n:
        if is_open:
            return not is_open, [html.I(className="fas fa-chevron-right")]
        else:
            return not is_open, [html.I(className="fas fa-chevron-down")]
    if is_open:
        return is_open, [html.I(className="fas fa-chevron-down")]
    else:
        return is_open, [html.I(className="fas fa-chevron-right")]

@app.callback(
    Output("save-as-collapse", "is_open"),
    Input("save-process-dropdown", "n_clicks"),
    State("save-as-collapse", "is_open"))
def toggle_expand_collapse(n, is_open):
    if n:
        return not is_open
    return is_open
    
@app.callback(
    Output("tooltip-content", "children"),
    Input("tab-state", "value"))
def render_tooltip(tab):
    if tab == 'discovery':
        return process_mining_explanation
    else:
        return conformance_explanation

@app.callback(Output('check-conformance-container', 'style'),
              Input('tab-state', 'value'))
def show_hide_button(tab):
    if tab == 'discovery':
        return {'display': 'None'}
    else:
        return {'display': 'block'}


small_loading_thread.join()

app.clientside_callback(
    """
    function placeholder(tab_state) {
        window.parent.WT1SVC.event('sol-open',
                {
                    'tab_name': tab_state,
                    'solution_name': 'process_mining'
                });
    };
    """,
    Output("dummy-event", "children"),
    Input("tab-state", "value")
)

app.clientside_callback(
    """
    function placeholder(n_clicks) {
        if (n_clicks > 0) {
            window.parent.WT1SVC.event('sol-action',
                {
                    'action_type': 'button',
                    'action_name': 'save_viz',
                    'solution_name': 'process_mining'
                });
        }
    };
    """,
    Output("dummy-event2", "children"),
    Input("modal-save-viz", "n_clicks")
)

app.clientside_callback(
    """
    function placeholder(n_clicks) {
        if (n_clicks > 0) {
            window.parent.WT1SVC.event('sol-action',
                {
                    'action_type': 'button',
                    'action_name': 'save_process',
                    'solution_name': 'process_mining'
                });
        }
    };
    """,
    Output("dummy-event3", "children"),
    Input("modal-save-process", "n_clicks")
)

app.clientside_callback(
    """
    function placeholder(n_clicks) {
        if (n_clicks > 0) {
            window.parent.WT1SVC.event('sol-action',
                {
                    'action_type': 'run_scenario_button',
                    'action_name': 'run_conformance_checks',
                    'solution_name': 'process_mining'
                });
        }
    };
    """,
    Output("dummy-event4", "children"),
    Input("conformance-button", "n_clicks")
)

layout_mining = html.Div([
    html.Script("""console.log("send_log");
                window.parent.WT1SVC.event('bs-process-mining-webapp-loaded');""", type="text/javascript"),
    dcc.Loading(id='loading', type='cube', fullscreen=True),
    dbc.Row([dbc.Col(children=[
                     filter_header(),
                     process_header_menu(),
                    html.Div(html.Button('Check Conformance', 
                                             id='conformance-button',
                                                    style=button_style_bottom_right),
                                 id='check-conformance-container',
                                 style={'display': 'None'}),
                     variants_container(),
                     filter_subtitle(),
                     dbc.Collapse(
                        base_filters(timeframe_df, start_end_df, attributes_values_df, attributes_columns, attributes_numeric, attribute_scales),
                        id="selectors-collapse",
                        is_open=True,
                     ),
                     html.Div([
                         exclude_subtitle(),
                         dbc.Collapse(
                            additional_filters(activity_count_df, events_filters, min_time_filter, self_transition),
                            id="filters-collapse",
                            is_open=True,
                     )],
                     id='additional-filters-zone')
                     
                     ], 
             style={'padding': '12px 24px',
                    'height': '100%',
                    'min-height': '100%',
                   'border-right': '1px solid',
                   'border-right-color': '#e3e4e4',
                   'overflow': 'scroll'},
                     md=3),
        dbc.Col([html.Div([dbc.Col(id='nb-cases-info', style={'float': 'left',
                                                             'margin': '8px',
                                                             'color': '#666666'}),
                         dbc.Col([html.Button('Discovery', id='discovery-tab', style=tab_button_style_active),
             html.Button('Conformance', id='conformance-tab', style=tab_button_style)],
            id='radio-buttons', style={'float': 'right',
                                       'margin': '8px',
                                       'padding': 0})],
                         style={'height': '48px', 
                                'margin-left': '-12px',
                                'border-bottom': '1px solid',
                                'border-bottom-color': '#e3e4e4',
                                'padding-top': '5px'}),
             dbc.Col(id='tab-content')
                         ], 
                md=9)], style={'height': '100%'}),
    dbc.Tooltip(process_mining_explanation,
            id='tooltip-content',
            target="question-tooltip",
            style={'color': '#bbbbbb',
                  'font-size': '10px'}),
    html.Div(id='dummy-event'),
    html.Div(id='dummy-event2'),
    html.Div(id='dummy-event3'),
    html.Div(id='dummy-event4'),
    dcc.Store(id='filters'),
    dcc.Store(id='unique-nodes'),
    dcc.Store(id='unique-nodes-individual'),
    dcc.Store(id='old-variants-filter', data=10),
    dcc.Store(id='saved-processes-value'),
    dcc.Store(id='dummy-input', data=1),
    dcc.Store(id='dummy-output', data=1),
    dcc.Store(id='tab-state', data='discovery'),
    dcc.Store(id='time-freq-state', data='frequency'),
    dcc.Store(id='time-freq-individual-state', data='time'),
    dcc.Store(id='list-chart-state', data='table')],
             style={'font-size': '12px', 'font-family': 'Helvetica Neue', 'height': '100%'})
logging.info("Current step: layout_mining, done")

conformance_check_filters = dbc.Col(base_filters(timeframe_df, start_end_df, attributes_values_df, attributes_columns, attributes_numeric, attribute_scales), 
                                   style={'margin-left': 0, 
                                          'border-right': '1px solid',
                                          'border-right-color': '#e3e4e4'})

# build your Dash app
app.layout = layout_mining